import urwid
from polka_curses.config import Palette as p


class BooksPage(urwid.ListBox):
    def __init__(self, books):
        self.books = books
        widgets = [BookItem(b) for b in self.books]
        super().__init__(urwid.SimpleFocusListWalker(widgets))

    def get_focused_book(self):
        widget, _ = self.get_focus()
        return widget.book


class BookItem(urwid.WidgetWrap):
    def __init__(self, book):
        self.book = book
        body = urwid.Text(self.body)
        body = urwid.Padding(body, left=1, right=1)
        item = urwid.LineBox(body, self.header, urwid.LEFT)
        super().__init__(urwid.AttrMap(item, p.frame.name, p.frame_focus.name))

    @property
    def header(self):
        return self.book.title

    @property
    def body(self):
        authors = ", ".join(self.book.authors)
        sy, ey = self.book.year
        years = f"{sy or ''}{'-' if sy and ey else ''}{ey or ''}"
        return f"{authors}{', ' if authors else ''}{years}"

    def selectable(self):
        return True



import urwid
from urwid import LEFT, SPACE


class ScrollableText(urwid.Text):
    def __init__(self, markup, align=LEFT, wrap=SPACE, layout=None, offset=1):
        self.offset = offset
        self._cache_trim_top = 0
        super().__init__(markup)

    def selectable(self):
        return True

    def render(self, size, focus=False):
        maxcol, maxrow = size

        super_canvas = super().render((maxcol,), focus)
        final_canvas = urwid.CompositeCanvas(super_canvas)

        cols, rows = final_canvas.cols(), final_canvas.rows()

        if cols <= maxcol:
            width = maxcol - cols
            if width > 0:
                final_canvas.pad_trim_left_right(0, width)

        if rows <= maxrow:
            height = maxrow - rows
            if height > 0:
                final_canvas.pad_trim_top_bottom(0, height)

        if cols <= maxcol and rows <= maxrow:
            return final_canvas

        trim_top = self._cache_trim_top
        trim_end = rows - maxrow - trim_top
        trim_right = cols - maxcol

        if trim_top > 0:
            final_canvas.trim(trim_top)

        if trim_end > 0:
            final_canvas.trim_end(trim_end)

        if trim_right > 0:
            final_canvas.pad_trim_left_right(0, -trim_right)

        return final_canvas

    def keypress(self, size, key):
        if key == "up":
            self._scroll_up(size)
        elif key == "down":
            self._scroll_down(size)
        else:
            return key

    def _scroll_up(self, size):
        new_trim_top = self._cache_trim_top - self.offset
        self._cache_trim_top = self._ensure_bounds(size, new_trim_top)
        self._invalidate()

    def _scroll_down(self, size):
        new_trim_top = self._cache_trim_top + self.offset
        self._cache_trim_top = self._ensure_bounds(size, new_trim_top)
        self._invalidate()

    def _ensure_bounds(self, size, new_trim_top):
        maxcol, maxrow = size
        rows = self.rows((maxcol,))
        return max(0, min(rows - maxrow, new_trim_top))



import urwid
from urwid import CENTER

from polka_curses.config import Palette as p


class TabBar(urwid.WidgetWrap):
    TABS = ["КНИГИ", " • ", "СПИСКИ", " • ", "ЭКСПЕРТЫ", " • ", "ПОИСК"]

    def __init__(self, index):
        self.index = index
        super().__init__(self._build())

    def _build(self):
        text = self.tabs
        text[self.i] = (p.header_focus.name, text[self.i])
        text = urwid.Text(text, align=CENTER)
        return urwid.AttrMap(text, p.header.name)

    def update(self):
        self._w = self._build()

    @property
    def tabs(self):
        return self.TABS.copy()

    @property
    def i(self):
        """The correct index that points out to the current tab."""
        return self.index * 2

    def _is_valid_index(self, index):
        return index >= 0 and index <= (len(self.tabs) // 2)

    def set_tab(self, index):
        if self._is_valid_index(index):
            self.index = index
            self.update()
        else:
            raise IndexError(f"tab bar index out of range")

    def get_current_tab(self):
        return self.tabs[self.i]

    def is_last_index(self):
        return self.index == (len(self.tabs) // 2)

    def is_first_index(self):
        return self.index == 0


class TitleHeader(urwid.WidgetWrap):
    def __init__(self, text):
        self.text = text
        super().__init__(self.build())

    def build(self):
        text = urwid.Text(self.text, align=CENTER)
        return urwid.AttrMap(text, p.header.name)

    @classmethod
    def init_for_book(cls, book):
        authors = ", ".join(book.authors) or None
        authors = "(" + authors + ")" if authors else ""
        authors = authors.upper()
        title = book.title.upper()
        return cls(f"{title} {authors}")

    @classmethod
    def init_for_list(cls, list_):
        title = f"{list_.title.upper()} - {list_.short_description.upper()}"
        return cls(title)

    @classmethod
    def init_for_expert(cls, expert):
        return cls(expert.name.upper())



import webbrowser
from collections import deque
from functools import wraps

import urwid

from .config import Mode, Palette, help_string_for
from .handler import InputHandler
from .model import Model
from .view import View


def update(func):
    """Updates the screen."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        self.loop.draw_screen()

    return wrapper


class ViewController:
    """The core of the program. It have methods that give commands to
    the `View` and have access to the Polka API through the `Model`."""

    def __init__(self):
        self.mode = Mode.LOADING_PAGE
        self.last_mode = deque()
        self.model = Model()
        self.view = View()
        self.loading = 10
        self.input_handler = InputHandler(self)

    def run(self):
        """Start the main loop."""
        self.loop = urwid.MainLoop(self.view)
        self.loop.handle_mouse = False
        self.loop.unhandled_input = self.input_handler.handle
        self.loop.screen.reset_default_terminal_palette()
        self.loop.screen.set_terminal_properties(256, False)
        self.loop.screen.register_palette(Palette.as_list())
        self.set_loading_page()
        self.model.get_all()
        self.loop.run()

    def exit(self):
        """Stop the main loop."""
        raise urwid.ExitMainLoop()

    def set_loading_page(self):
        callback = self.update_loading_page
        self.remove_handler = self.loop.set_alarm_in(0.2, callback)

    def remove_loading_page(self):
        self.loop.remove_alarm(self.remove_handler)

    @update
    def update_loading_page(self, *args, **kwargs):
        if self.model.is_loaded:
            self.remove_loading_page()
            rmsg = help_string_for(Mode.BOOKS_PAGE)
            self.view.init(self.model.books, rmsg=rmsg)
            self.set_mode(Mode.BOOKS_PAGE)
        else:
            self.view.update_loading_page()
            self.set_loading_page()

    def set_mode(self, mode, save_last=False):
        if save_last:
            self.last_mode.append(self.mode)
        self.mode = mode

    def get_mode(self):
        return self.mode

    def get_last_mode(self):
        return self.last_mode.pop() if self.last_mode else None

    def move_right(self):
        """Go to the next tab on the right."""
        tab_name = self.view.focus_next_tab()
        self.draw_body_by_tab_name(tab_name)

    def move_left(self):
        """Go to the previous tab on the left."""
        tab_name = self.view.focus_previous_tab()
        self.draw_body_by_tab_name(tab_name)

    @update
    def draw_body_by_tab_name(self, tab_name):
        mode = Mode.get(tab_name)
        help_string = help_string_for(mode)
        if mode == Mode.BOOKS_PAGE:
            self.view.draw_books(self.model.books, rmsg=help_string)
        elif mode == Mode.LISTS_PAGE:
            self.view.draw_lists(self.model.lists, rmsg=help_string)
        elif mode == Mode.EXPERTS_PAGE:
            self.view.draw_experts(self.model.experts, rmsg=help_string)
        else:
            self.view.draw_search(rmsg=help_string)
        self.set_mode(mode, save_last=True)

    @update
    def show_search_results(self):
        """If the results exist, show them, and if not, show a new
        search page."""
        query = self.view.get_search_query()
        results = self.model.search(query)
        if results:
            mode = Mode.SEARCH_RESULTS_PAGE
            help_string = help_string_for(mode)
            self.view.show_search_results(results, rmsg=help_string)
        else:
            mode = Mode.SEARCH_PAGE
            help_string = help_string_for(mode)
            self.view.draw_search("Нет результатов", help_string)
        self.set_mode(mode, save_last=True)

    @update
    def show_search(self):
        """Show search page."""
        mode = Mode.SEARCH_PAGE
        help_string = help_string_for(mode)
        self.view.draw_search(rmsg=help_string)
        self.set_mode(mode, save_last=True)

    @update
    def open_search_result(self):
        result = self.view.get_focused_search_result()
        if self.model.is_book(result):
            if result.has_article:
                mode = Mode.BOOK_PAGE
                rmsg = help_string_for(mode)
                self.view.draw_book(result, rmsg=rmsg)
            else:
                self.view.write_error_the_book_has_no_page(result)
                return None
        elif self.model.is_list(result):
            mode = Mode.LIST_PAGE
            rmsg = help_string_for(mode)
            self.view.draw_list(result, rmsg=rmsg)
        else:
            mode = Mode.EXPERT_PAGE
            rmsg = help_string_for(mode)
            self.view.draw_expert(result, rmsg=rmsg)
        self.set_mode(mode, save_last=True)

    @update
    def open_book(self):
        book = self.view.get_focused_book()
        if self.model.book_has_article(book):
            mode = Mode.BOOK_PAGE
            rmsg = help_string_for(mode)
            self.view.draw_book(book, rmsg=rmsg)
            self.set_mode(mode, save_last=True)
        else:
            self.view.write_error_the_book_has_no_page(book)

    @update
    def open_list(self):
        mode = Mode.LIST_PAGE
        list_ = self.view.get_focused_list()
        rmsg = help_string_for(mode)
        self.view.draw_list(list_, rmsg=rmsg)
        self.set_mode(mode, save_last=True)

    @update
    def open_expert(self):
        mode = Mode.EXPERT_PAGE
        expert = self.view.get_focused_expert()
        rmsg = help_string_for(mode)
        self.view.draw_expert(expert, rmsg=rmsg)
        self.set_mode(mode, save_last=True)

    @update
    def show_list_description(self):
        mode = Mode.LIST_DESCRIPTION_PAGE
        rmsg = help_string_for(mode)
        self.view.draw_list_description(rmsg=rmsg)
        self.set_mode(mode, save_last=True)

    @update
    def open_previous(self):
        mode = self.get_last_mode()
        if mode:
            self.view.draw_previous()
            self.set_mode(mode)

    def open_in_browser(self):
        if self.mode == Mode.BOOK_PAGE:
            url = self.view.get_book_article_url()
        elif self.mode == Mode.LIST_PAGE:
            url = self.view.get_list_article_url()
        elif self.mode == Mode.EXPERT_PAGE:
            url = self.view.get_expert_article_url()
        webbrowser.open(url)



import urwid
from urwid import PACK

from .books_page import BooksPage


class ExpertPage(urwid.WidgetWrap):
    def __init__(self, expert):
        self.expert = expert
        self.is_two_columns = expert.favorites and expert.wrote_about
        self.favorites = self.build_favorites()  # type: BooksPage
        self.wrote_about = self.build_wrote_about()  # type: BooksPage
        self.columns = {0: self.favorites, 1: self.wrote_about}
        super().__init__(self.build())

    def build(self):
        if self.is_two_columns:
            favorites = urwid.LineBox(self.favorites, title="ВЫБОР ЭКСПЕРТА")
            wrote_about = urwid.LineBox(self.wrote_about, title="СТАТЬИ")
            self.container = urwid.Columns([favorites, wrote_about], 1, 0)
        elif self.favorites:
            favorites = urwid.LineBox(self.favorites, title="ВЫБОР ЭКСПЕРТА")
            self.container = favorites
        elif self.wrote_about:
            wrote_about = urwid.LineBox(self.wrote_about, title="СТАТЬИ")
            self.container = wrote_about
        else:
            self.container = urwid.Filler(urwid.Text(""))
        return urwid.Pile([(PACK, self.description), self.container])

    def build_favorites(self):
        if self.expert.favorites:
            return BooksPage(self.expert.favorites)
        else:
            return None

    def build_wrote_about(self):
        if self.expert.wrote_about:
            return BooksPage(self.expert.wrote_about)
        else:
            return None

    @property
    def description(self):
        desc = urwid.Text(self.expert.description)
        desc = urwid.Padding(desc, left=1, right=1)
        return urwid.LineBox(desc)

    def get_focused_book(self):
        column = self.get_focused_column()
        return column.get_focused_book()

    def get_focused_column(self):
        if self.is_two_columns:
            return self.columns[self.container.focus_position]
        return self.favorites or self.wrote_about



from enum import Enum
from typing import NamedTuple, List, Set


class Mode(Enum):
    """The state in which the program is running."""

    BOOKS_PAGE = "КНИГИ"
    LISTS_PAGE = "СПИСКИ"
    EXPERTS_PAGE = "ЭКСПЕРТЫ"
    SEARCH_PAGE = "ПОИСК"
    SEARCH_RESULTS_PAGE = "РЕЗУЛЬТАТЫ ПОИСКА"
    BOOK_PAGE = "СТАТЬЯ"
    LIST_PAGE = "СПИСОК"
    EXPERT_PAGE = "ЭКСПЕРТ"
    LIST_DESCRIPTION_PAGE = "ИНФОРМАЦИЯ О СПИСКЕ"
    LOADING_PAGE = "СТРАНИЦА ЗАГРУЗКИ"

    @classmethod
    def get(cls, raw_value):
        for mode in cls:
            if mode.value == raw_value:
                return mode

    @classmethod
    def pages(cls):
        return {cls.BOOK_PAGE, cls.LIST_PAGE, cls.EXPERT_PAGE}

    @classmethod
    def tabs(cls):
        return {
            cls.BOOKS_PAGE,
            cls.LISTS_PAGE,
            cls.EXPERTS_PAGE,
            cls.SEARCH_PAGE,
            cls.SEARCH_RESULTS_PAGE,
        }


WHITE = "#fff"
BLACK = "#000"
BLUE = "#00f"
YELLOW = "#ff0"
RED = "#f00"
BOLD = "bold"


class PaletteItem(NamedTuple):
    name: str
    fore: str = ""
    back: str = ""
    mono: str = ""
    fore_high: str = ""
    back_high: str = ""


class Palette(Enum):
    header = PaletteItem("header", fore_high=WHITE, back_high=BLUE)
    header_focus = PaletteItem("header_focus", fore_high=WHITE, back_high=RED)
    footer = PaletteItem("footer", fore_high=WHITE, back_high=BLUE)
    footer_error = PaletteItem("footer_error", fore_high=WHITE, back_high=RED)
    frame = PaletteItem("frame", fore_high=BLACK, back_high=WHITE)
    frame_focus = PaletteItem("frame_focus", fore_high=BLACK, back_high=YELLOW)
    frame_bold = PaletteItem(
        "frame_bold", fore_high=f"{BOLD},{BLACK}", back_high=WHITE
    )
    frame_bold_focus = PaletteItem(
        "frame_bold_focus", fore_high=f"{BOLD},{BLACK}", back_high=YELLOW
    )

    @classmethod
    def as_list(cls):
        return [p.value for p in cls]


class Key(NamedTuple):
    """Represents a keyboard input sent by a user’s terminal."""

    name: str
    description: str
    buttons: List[str]
    action: str
    modes: Set[Mode]
    hidden: bool = False


KEYS = [
    Key(
        name="Left",
        description="ВЛЕВО",
        buttons=["left"],
        action="move_left",
        modes=Mode.tabs(),
        hidden=True,
    ),
    Key(
        name="Right",
        description="ВПРАВО",
        buttons=["right"],
        action="move_right",
        modes=Mode.tabs(),
        hidden=True,
    ),
    Key(
        name="Space",
        buttons=[" "],
        action="show_list_description",
        description="ОПИСАНИЕ",
        modes={Mode.LIST_PAGE},
    ),
    Key(
        name="Space",
        buttons=[" "],
        action="open_previous",
        description="ЗАКРЫТЬ",
        modes={Mode.LIST_DESCRIPTION_PAGE},
    ),
    Key(
        name="Enter",
        buttons=["enter"],
        action="open_book",
        description="ОТКРЫТЬ",
        modes={Mode.BOOKS_PAGE, Mode.LIST_PAGE, Mode.EXPERT_PAGE},
    ),
    Key(
        name="Enter",
        buttons=["enter"],
        action="open_list",
        description="ОТКРЫТЬ",
        modes={Mode.LISTS_PAGE},
    ),
    Key(
        name="Enter",
        buttons=["enter"],
        action="open_expert",
        description="ОТКРЫТЬ",
        modes={Mode.EXPERTS_PAGE},
    ),
    Key(
        name="Enter",
        description="ИСКАТЬ",
        buttons=["enter"],
        action="show_search_results",
        modes={Mode.SEARCH_PAGE},
    ),
    Key(
        name="Tab",
        buttons=["tab"],
        action="open_in_browser",
        description="ОТКРЫТЬ В БРАУЗЕРЕ",
        modes=Mode.pages(),
    ),
    Key(
        name="Enter",
        buttons=["enter"],
        action="open_search_result",
        description="ОТКРЫТЬ",
        modes={Mode.SEARCH_RESULTS_PAGE},
    ),
    Key(
        name="Space",
        action="show_search",
        buttons=[" "],
        description="НОВЫЙ ПОИСК",
        modes={Mode.SEARCH_RESULTS_PAGE},
    ),
    Key(
        name="Esc",
        buttons=["esc"],
        action="open_previous",
        description="ЗАКРЫТЬ",
        modes=Mode.pages(),
    ),
    Key(
        name="Esc",
        description="ВЫХОД",
        buttons=["esc"],
        action="exit",
        modes=Mode.tabs().union({Mode.LOADING_PAGE}),
    ),
]


def get_keys_by_mode(mode):
    """Filtres `KEYS` by mode."""
    return (key for key in KEYS if mode in key.modes)


def get_key_action(mode, button):
    """Extracts the controller method name from a Key."""
    for item in get_keys_by_mode(mode):
        if button in item.buttons:
            return item.action
    return None


def help_string_for(mode):
    """Returns a string like `"ВЫХОД (Esc), ОТКРЫТЬ (Enter), ..."`."""
    keys = get_keys_by_mode(mode)
    keys = [f"{k.description} ({k.name})" for k in keys if not k.hidden]
    keys = ", ".join(keys)
    return keys



import urwid
from urwid import LEFT, RIGHT, PACK

from polka_curses.config import Palette as p


class StatusBar(urwid.WidgetWrap):
    def __init__(self, left="", right=""):
        lcol = urwid.Text(left, align=LEFT)
        rcol = urwid.Text(right, align=RIGHT)
        cols = urwid.Columns([lcol, (PACK, rcol)], dividechars=1)
        super().__init__(urwid.AttrMap(cols, p.footer.name))

    @property
    def cols(self):
        return self._w.original_widget.base_widget.contents

    @property
    def left(self):
        return self.cols[0][0].base_widget.text

    @property
    def right(self):
        return self.cols[1][0].base_widget.text

    def set_left(self, text, error=False):
        if error:
            text = urwid.AttrMap(urwid.Text(text), p.footer_error.name)
        else:
            text = urwid.AttrMap(urwid.Text(text), p.footer.name)
        cols = self._w.original_widget.base_widget
        cols.contents[0] = (text, cols.contents[0][1])

    def set_right(self, text):
        text = urwid.AttrMap(urwid.Text(text), p.footer.name)
        cols = self._w.original_widget.base_widget
        cols.contents[1] = (text, cols.contents[1][1])

    def clear_left(self):
        self.set_left("")

    def clear_right(self):
        self.set_right("")



import urwid

from polka_curses.config import Palette as p


class ListsPage(urwid.ListBox):
    def __init__(self, lists):
        self.lists = lists
        widgets = [ListItem(l) for l in self.lists]
        super().__init__(urwid.SimpleFocusListWalker(widgets))

    def get_focused_list(self):
        widget, _ = self.get_focus()
        return widget.list_


class ListItem(urwid.WidgetWrap):
    def __init__(self, list_):
        self.list_ = list_
        body = urwid.Text(self.body)
        body = urwid.Padding(body, left=1, right=1)
        item = urwid.LineBox(body, self.header, urwid.LEFT)
        super().__init__(urwid.AttrMap(item, p.frame.name, p.frame_focus.name))

    @property
    def header(self):
        return self.list_.title

    @property
    def body(self):
        return self.list_.short_description

    def selectable(self):
        return True



import urwid
from urwid import LEFT

from polka_curses.config import Palette as p


class ExpertsPage(urwid.ListBox):
    def __init__(self, experts):
        self.experts = experts
        widgets = [ExpertItem(e) for e in self.experts]
        super().__init__(urwid.SimpleFocusListWalker(widgets))

    def get_focused_expert(self):
        widget, _ = self.get_focus()
        return widget.expert


class ExpertItem(urwid.WidgetWrap):
    def __init__(self, expert):
        self.expert = expert
        body = urwid.Text(self.body)
        body = urwid.Padding(body, left=1, right=1)
        item = urwid.LineBox(body, self.header, LEFT)
        super().__init__(urwid.AttrMap(item, p.frame.name, p.frame_focus.name))

    @property
    def header(self):
        return self.expert.name

    @property
    def body(self):
        return self.expert.credit

    def selectable(self):
        return True



import urwid
from urwid import CENTER, LEFT, MIDDLE

from polka_curses.config import Palette as p

from .widgets.scrollable_text import ScrollableText


class ListPage(urwid.ListBox):
    def __init__(self, list_):
        self.list_ = list_
        widgets = [BookInListItem(b) for b in self.list_.books]
        super().__init__(urwid.SimpleFocusListWalker(widgets))

    def get_focused_book(self):
        widget, _ = self.get_focus()
        return widget.book


class BookInListItem(urwid.WidgetWrap):
    def __init__(self, book):
        self.book = book
        body = urwid.Padding(self.body, left=1, right=1)
        item = urwid.LineBox(body, self.header, LEFT)
        super().__init__(urwid.AttrMap(item, p.frame.name, p.frame_focus.name))

    @property
    def header(self):
        return self.book.title

    @property
    def body(self):
        authors = ", ".join(self.book.authors)
        sy, ey = self.book.year
        years = f"{sy or ''}{'-' if sy and ey else ''}{ey or ''}"
        sep = ", " if authors else ""
        desc = self.book.description
        return urwid.Text(f"\n{authors}{sep}{years}\n\n{desc}")

    def selectable(self):
        return True


class ListDescription(urwid.WidgetWrap):
    def __init__(self, list_, bg, **kwargs):
        self.list_ = list_
        self.bg = bg
        super().__init__(self.build())

    def build(self):
        w = ScrollableText(self.list_.description)
        w = urwid.Padding(w, left=2, right=2, align=CENTER)
        w = urwid.LineBox(w, "📜", CENTER)
        return urwid.Overlay(w, self.bg, CENTER, 65, MIDDLE, 15)



import urwid


class LoadingPage(urwid.WidgetWrap):
    def __init__(self):
        self.emoji = ["🌍", "🌏", "🌎"]
        w = urwid.Text("🌍", align=urwid.CENTER)
        w = urwid.Filler(w)
        super().__init__(w)

    @property
    def text(self):
        return self._w.base_widget.text

    @text.setter
    def text(self, value):
        self._w.base_widget.set_text(value)

    def update(self):
        i = self.emoji.index(self.text)  # the current emoji index
        next_i = (i + 1) % len(self.emoji)  # the next emoji index
        self.text = self.emoji[next_i]



from builtins import property

import urwid

from polka_curses.views.widgets.scrollable_text import ScrollableText
from polka_curses.config import Palette as p


class BookPage(urwid.WidgetWrap):
    def __init__(self, book):
        self.book = book
        super().__init__(self.build())

    @property
    def questions(self):
        return [Question(q, a) for q, a, _ in self.book.questions]

    def build(self):
        lcol = urwid.LineBox(Questions(self.questions, self.change_answer))
        rcol = Answer("")
        self.cols = urwid.Columns([(35, lcol), rcol], focus_column=0)
        self.questions_column.update_answer()
        return self.cols

    def change_answer(self, answer):
        options = self.cols.contents[1][1]
        answer = urwid.LineBox(urwid.Padding(Answer(answer), left=1, right=1))
        answer = urwid.AttrMap(answer, p.frame.name, p.frame_focus.name)
        self.cols.contents[1] = (answer, options)

    @property
    def questions_column(self):
        return self.cols.contents[0][0].base_widget

    @property
    def answer_column(self):
        return self.cols.contents[1][0].base_widget


class Question(urwid.WidgetWrap):
    def __init__(self, question, answer):
        self.question = question
        self.answer = answer
        super().__init__(self.build())

    def build(self):
        borders = {
            "tlcorner": "",
            "tline": " ",
            "lline": " ",
            "trcorner": "",
            "blcorner": "",
            "rline": " ",
            "bline": " ",
            "brcorner": "",
        }
        w = urwid.Text(self.question.upper())
        w = urwid.LineBox(w, **borders)
        return urwid.AttrMap(w, p.frame_bold.name, p.frame_bold_focus.name)

    def selectable(self):
        return True


class Questions(urwid.ListBox):
    def __init__(self, questions, func_for_changing_answer):
        self.questions = questions
        self.change_answer = func_for_changing_answer
        super().__init__(urwid.SimpleFocusListWalker(self.questions))

    def change_focus(self, *args, **kwargs):
        super().change_focus(*args, **kwargs)
        self.update_answer()

    def update_answer(self):
        self.change_answer(self.focus.answer)


class Answer(ScrollableText):
    pass



import urwid
from urwid import CENTER, RELATIVE, TOP, LEFT

from polka_curses.config import Palette as p


RELATIVE_55 = (RELATIVE, 55)
RELATIVE_90 = (RELATIVE, 90)


class SearchResultsExseption(Exception):
    pass


class SearchPage(urwid.WidgetWrap):
    def __init__(self):
        self.search = urwid.Edit()
        bg = urwid.SolidFill("⢻")
        w = urwid.Filler(urwid.LineBox(self.search, "ПОИСК 🔎", CENTER))
        w = urwid.Overlay(w, bg, CENTER, RELATIVE_55, TOP, 3, top=2)
        super().__init__(w)

    def get_search_query(self):
        return self.search.edit_text


class SearchResultsPage(urwid.WidgetWrap):
    def __init__(self, results):
        self.results = SearchResults(results)
        bg = urwid.SolidFill("⢻")
        w = urwid.LineBox(self.results, "ПОИСК 🔎", CENTER)
        w = urwid.Overlay(w, bg, CENTER, RELATIVE_55, TOP, RELATIVE_90, top=2)
        super().__init__(w)

    def get_focused_result(self):
        return self.results.get_focused_result()


class SearchResults(urwid.ListBox):
    def __init__(self, results):
        self.results = [SearchResultItem(r) for r in results]
        super().__init__(urwid.SimpleFocusListWalker(self.results))

    def get_focused_result(self):
        if self.results:
            w, _ = self.get_focus()
            return w.result.object
        else:
            raise SearchResultsExseption("there is no results")


class SearchResultItem(urwid.WidgetWrap):
    def __init__(self, result):
        self.result = result
        body = urwid.Text(self.body)
        body = urwid.Padding(body, left=1, right=1)
        item = urwid.LineBox(body, self.header, LEFT)
        super().__init__(urwid.AttrMap(item, p.frame.name, p.frame_focus.name))

    @property
    def header(self):
        return self.result.title

    @property
    def body(self):
        return f"{self.result.description}..."

    def selectable(self):
        return True



from .config import get_key_action


class InputHandler:
    """Maps user input to calls to `Controller` functions."""

    def __init__(self, controller):
        self.controller = controller

    def do(self, action):
        for method in dir(self.controller):
            if method == action:
                getattr(self.controller, action)()

    def handle(self, key):
        action = get_key_action(self.controller.mode, key)
        self.do(action)



from collections import deque
from functools import wraps

import urwid

from polka_curses.views.search_page import SearchResultsPage

from .config import Palette as p
from .views.book_page import BookPage
from .views.books_page import BooksPage
from .views.expert_page import ExpertPage
from .views.experts_page import ExpertsPage
from .views.list_page import ListDescription, ListPage
from .views.lists_page import ListsPage
from .views.loading_page import LoadingPage
from .views.search_page import SearchPage, SearchResultsPage
from .views.status_bar import StatusBar
from .views.tab_bar import TabBar, TitleHeader


def save_previous(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self.previous_bodies.append(self.body)
        self.previous_footers.append(self.footer)
        self.previous_headers.append(self.header)
        return func(self, *args, **kwargs)

    return wrapper


def in_pages(*pages):
    """A decorator for View's methods. If a method was called and the
    current body class is not in `pages` the decorator raises an error.
    Use it only if specific objects are expected inside the method."""

    def inner(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not isinstance(self.body, tuple(pages)):
                raise ViewError.wrong_body(self.body, pages)
            return func(self, *args, **kwargs)

        return wrapper

    return inner


class ViewError(Exception):
    @classmethod
    def wrong_body(cls, got, expected):
        got = got.__class__.__name__
        expected = [e.__name__ for e in expected]
        return cls(f"Wrong body object: {got}. Expected: {expected}")


class View(urwid.WidgetWrap):
    """It is intended for managing the UI views."""

    def __init__(self):
        self.frame = urwid.Frame(LoadingPage())
        self.previous_headers = deque()
        self.previous_bodies = deque()
        self.previous_footers = deque()
        super().__init__(urwid.AttrMap(self.frame, p.frame.name))

    @in_pages(LoadingPage)
    def update_loading_page(self):
        self.body.update()

    def init(self, books, lmsg="", rmsg=""):
        self.set_body(BooksPage(books))
        self.set_header(TabBar(0))
        self.set_footer(StatusBar(lmsg, rmsg))

    @property
    def header(self):
        return self.frame.header

    @property
    def body(self):
        return self.frame.body

    @property
    def footer(self):
        return self.frame.footer

    def set_body(self, body):
        self.frame.set_body(body)

    def set_header(self, header):
        self.frame.set_header(header)

    def set_footer(self, footer):
        self.frame.set_footer(footer)

    def draw_books(self, books, lmsg="", rmsg=""):
        """Replaces the body with a `BooksPage` object and the frame
        footer with a new `StatusBar` object."""
        self.set_body(BooksPage(books))
        self.set_footer(StatusBar(lmsg, rmsg))

    def draw_lists(self, lists, lmsg="", rmsg=""):
        """Replaces the body with a `ListsPage` object and the frame
        footer with a new `StatusBar` object."""
        self.set_body(ListsPage(lists))
        self.set_footer(StatusBar(lmsg, rmsg))

    def draw_experts(self, experts, lmsg="", rmsg=""):
        """Replaces the body with an `ExpertsPage` object and the frame
        footer with a new `StatusBar` object."""
        self.set_body(ExpertsPage(experts))
        self.set_footer(StatusBar(lmsg, rmsg))

    def draw_search(self, lmsg="", rmsg=""):
        """Replaces the body with a `SearchPage` object and the frame
        footer with a new `StatusBar` object."""
        self.set_body(SearchPage())
        self.set_footer(StatusBar(lmsg, rmsg))

    def show_search_results(self, results, lmsg="", rmsg=""):
        """Replaces the body with a `SearchResultsPage` object and
        the frame footer with a new `StatusBar` object."""
        self.set_body(SearchResultsPage(results))
        self.set_footer(StatusBar(lmsg, rmsg))

    @in_pages(SearchPage)
    def get_search_query(self):
        """Returns the search query from a `SearchResultsPage` object"""
        return self.body.get_search_query()

    @in_pages(BooksPage, ListsPage, ExpertsPage, SearchPage, SearchResultsPage)
    def focus_previous_tab(self):
        """Focuses the previous tab in the header."""
        if not self.header.is_first_index():
            self.header.set_tab(self.header.index - 1)
            self.footer.clear_left()
        return self.header.get_current_tab()

    @in_pages(BooksPage, ListsPage, ExpertsPage, SearchPage, SearchResultsPage)
    def focus_next_tab(self):
        """Focuses the next tab in the header."""
        if not self.header.is_last_index():
            self.header.set_tab(self.header.index + 1)
            self.footer.clear_left()
        return self.header.get_current_tab()

    @in_pages(ListsPage)
    def get_focused_list(self):
        return self.body.get_focused_list()

    @in_pages(BooksPage, ListPage, ExpertPage)
    def get_focused_book(self):
        return self.body.get_focused_book()

    @in_pages(ExpertsPage)
    def get_focused_expert(self):
        return self.body.get_focused_expert()

    @in_pages(SearchResultsPage)
    def get_focused_search_result(self):
        return self.body.get_focused_result()

    @save_previous
    def draw_book(self, book, lmsg="", rmsg=""):
        """Replaces the body with a `BookPage` object and
        the frame footer with a new `StatusBar` object."""
        self.set_body(BookPage(book))
        self.set_footer(StatusBar(lmsg, rmsg))
        self.set_header(TitleHeader.init_for_book(book))

    @save_previous
    def draw_list(self, list_, lmsg="", rmsg=""):
        """Replaces the body with a `ListPage` object and
        the frame footer with a new `StatusBar` object."""
        self.set_body(ListPage(list_))
        self.set_footer(StatusBar(lmsg, rmsg))
        self.set_header(TitleHeader.init_for_list(list_))

    @save_previous
    def draw_expert(self, expert, lmsg="", rmsg=""):
        """Replaces the body with an `ExpertPage` object and
        the frame footer with a new `StatusBar` object."""
        self.set_body(ExpertPage(expert))
        self.set_footer(StatusBar(lmsg, rmsg))
        self.set_header(TitleHeader.init_for_expert(expert))

    @save_previous
    @in_pages(ListPage)
    def draw_list_description(self, lmsg="", rmsg=""):
        self.set_body(ListDescription(self.body.list_, bg=self.body))
        self.set_footer(StatusBar(lmsg, rmsg))

    def write_error_the_book_has_no_page(self, book):
        text = f"Нет статьи на книгу «{book.title.upper()}»"
        self.footer.set_left(text, error=True)

    def draw_previous(self):
        if self.previous_bodies:
            self.set_body(self.previous_bodies.pop())
        if self.previous_footers:
            self.set_footer(self.previous_footers.pop())
            self.footer.clear_left()
        if self.previous_headers:
            self.set_header(self.previous_headers.pop())

    @in_pages(BookPage)
    def get_book_article_url(self):
        return self.body.book.url

    @in_pages(ListPage)
    def get_list_article_url(self):
        return self.body.list_.url

    @in_pages(ExpertPage)
    def get_expert_article_url(self):
        return self.body.expert.url



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['polka_curses', 'polka_curses.views', 'polka_curses.views.widgets']

package_data = \
{'': ['*']}

install_requires = \
['polka>=2.4,<3.0', 'urwid>=2.0,<3.0']

entry_points = \
{'console_scripts': ['polka = polka_curses.main:main']}

setup_kwargs = {
    'name': 'polka-curses',
    'version': '1.1.3',
    'description': 'The curses interface for the Polka website (https://polka.academy/)',
    'long_description': '![](./images/install_and_usage.gif)\n',
    'author': 'Dima Koskin',
    'author_email': 'dmksknn@gmail.com',
    'url': 'https://github.com/dmkskn/polka_curses/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from functools import wraps
from threading import Thread

import polka


def threaded(func):
    @wraps(func)
    def async_func(*args, **kwargs):
        thread = Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return async_func


class Model:
    """Access to the Polka API"""

    def __init__(self):
        self._books = []
        self._lists = []
        self._experts = []
        self.is_loaded = False

    @property
    def books(self):
        """Gets list of books. Each book is an instance of polka.Book
        class."""
        if not self._books:
            self._books = polka.books()
        return self._books

    @property
    def lists(self):
        """Gets list of compilations. Each compilations is an instance
        of polka.Compilation class."""
        if not self._lists:
            self._lists = polka.lists()
        return self._lists

    @property
    def experts(self):
        """Gets list of experts. Each expert is an instance
        of polka.Pundit class."""
        if not self._experts:
            self._experts = polka.pundits()
        return self._experts

    def search(self, query):
        """Search for the `query`. Each expert is a tuple `(title,
        description, object)`."""
        return polka.search(query)

    def is_book(self, obj):
        return isinstance(obj, polka.Book)

    def is_list(self, obj):
        return isinstance(obj, polka.Compilation)

    def is_expert(self, obj):
        return isinstance(obj, polka.Pundit)

    def book_has_article(self, book: polka.Book):
        return book.has_article

    @threaded
    def get_all(self):
        """Preload all data"""
        _ = self.books
        _ = self.lists
        _ = self.experts
        self.is_loaded = True
